@@ -1,7 +1,7 @@
 /*
- * ScanSSH - simple SSH version scanner V1.3
+ * ScanSSH - simple SSH version scanner V1.4
  *
- * Copyright 2000 Niels Provos <provos@citi.umich.edu>
+ * Copyright 2000-2001 Niels Provos <provos@citi.umich.edu>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -46,16 +46,59 @@
 #include <netdb.h>
 #include <pcap.h>
 #include <unistd.h>
+#include <md5.h>
 
+#include "config.h"
+#ifdef HAVE_FDMASK_IN_SELECT
+#include <sys/select.h>
+#endif
+
+#include "scanssh.h"
 #include "exclude.h"
 #include "pcapu.h"
+#include "xmalloc.h"
 
 #ifndef howmany
 #define howmany(x,y)	(((x) + ((y) - 1)) / (y))
 #endif
 
-ssize_t atomicio(ssize_t (*)(), int, void *, size_t);
-int ipv4toa(char *, size_t, void *);
+struct address_node {
+	TAILQ_ENTRY (address_node) an_next;
+
+	sa_family_t an_type;
+	union {
+		struct in_addr ipv4;
+	} an_ipstart;
+#define an_ipv4start an_ipstart.ipv4
+
+	union {
+		struct in_addr ipv4;
+	} an_ipend;
+#define an_ipv4end an_ipend.ipv4
+
+	int an_bits;
+};
+
+struct generate {
+	TAILQ_ENTRY (generate) gen_next;
+
+	TAILQ_HEAD (an_list, address_node) gen_anqueue;
+
+	int gen_flags;
+
+	u_int32_t gen_seed;		/* Seed for PRNG */
+
+	u_int32_t gen_bits;
+	u_int32_t gen_start;
+	u_int32_t gen_iterate;
+	u_int32_t gen_end;
+	u_int32_t gen_current;
+	u_int32_t gen_n;		/* successful generations */
+	u_int32_t gen_max;
+};
+
+int populate(struct argument *, int);
+int next_address(struct generate *, struct in_addr *);
 
 /* Globals */
 int usepcap;
@@ -66,35 +109,13 @@ int rndexclude = 1;
 
 int commands[MAX_PROCESSES];
 int results[MAX_PROCESSES];
-int alarmed;
 
-#define SSHMAPVERSION	"SSH-1.0-SSH_Version_Mapper\n"
-#define MAXITER		10
-#define LONGWAIT	30
-#define SHORTWAIT	2
-#define CONNECTWAIT	5
-#define SYNWAIT		1
-#define SYNRETRIES	5
-#define MAXBUF		2048
-#define MAXSYNQUEUESZ	256
-#define MAXBURST	64
+struct argument *args;
 
+TAILQ_HEAD (gen_list, generate) genqueue;
 TAILQ_HEAD (queue_list, argument) readyqueue;
 TAILQ_HEAD (syn_list, argument) synqueue;
 
-struct argument {
-	TAILQ_ENTRY (argument) a_next;
-	struct timeval a_tv;
-	sa_family_t a_type;
-	int a_retry;		/* what a waste of memory */
-	u_int32_t a_seqnr;
-	union {
-		struct in_addr ipv4;
-	} ip;
-
-#define a_ipv4 ip.ipv4
-};
-
 #define synlist_hasspace()	(synqueuesz < MAXSYNQUEUESZ)
 #define synlist_empty()		(synqueuesz == 0)
 
@@ -186,147 +207,6 @@ sigchld_handler(int sig)
         errno = save_errno;
 }
 
-void
-sigalrm_handler(int sig)
-{
-        int save_errno = errno;
-	alarmed = 1;
-        errno = save_errno;
-}
-
-int
-scanhost(struct argument *arg, char *buf, size_t size)
-{
-	int res, sock, j;
-	struct addrinfo hints, *ai;
-	char firstline[255];
-	char ntop[NI_MAXHOST];
-	
-	switch (arg->a_type) {
-	case AF_INET:
-		ipv4toa(ntop, sizeof(ntop), &arg->a_ipv4);
-		break;
-	default:
-		strlcpy(buf, "<unsuppfamily>", size);
-		return (-1);
-	}
-
-	sock = socket(arg->a_type, SOCK_STREAM, 0);
-	if (sock == -1) {
-		strlcpy(buf, "<socketcreate>", size);
-		return (-1);
-	}
-
-	memset(&hints, 0, sizeof(hints));
-        hints.ai_family = AF_UNSPEC;
-        hints.ai_socktype = SOCK_STREAM;
-        if (getaddrinfo(ntop, "22", &hints, &ai) != 0) {
-		strlcpy(buf, "<getaddrinfo>", size);
-		return (-1);
-	}
-
-	alarm(CONNECTWAIT);
-	res = connect(sock, ai->ai_addr, ai->ai_addrlen);
-	alarm(0);
-	if (res == -1) {
-		close(sock);
-		switch (errno) {
-		case ETIMEDOUT:
-		case EINTR:
-#ifdef __linux__
-		case EHOSTUNREACH:
-#endif
-			strlcpy(buf, "<timeout>", size);
-			return (-1);
-		case ECONNREFUSED:
-			strlcpy(buf, "<refused>", size);
-			return (-1);
-		case ENETUNREACH:
-			strlcpy(buf, "<unreachable>", size);
-			return (-1);
-		default:
-			snprintf(buf, size, "<%s>", strerror(errno));
-			return (-1);
-		}
-	}
-	
-	/* Read other side\'s version identification. */
-        for (j = 0; j < MAXITER; j++) {
-		int i;
-                for (i = 0; i < size - 1; i++) {
-                        int len = -1;
-			if (!j && !i)
-				alarm (LONGWAIT);
-			else
-				alarm (SHORTWAIT);
-			while ((len = read(sock, &buf[i], 1)) == -1) {
-				if (len == -1 && errno == EINTR)
-					break;
-			}
-			alarm(0);
-                        if (len < 0 && !j) {
-				close(sock);
-				strlcpy(buf, "<timeout>", size);
-				return (-1);
-			} else if (len != 1 && !j) {
-				close(sock);
-                                strlcpy(buf, "<closed>", size);
-				return (-1);
-			} else if (len != 1) {
-				j = MAXITER;
-				break;
-			}
-			
-                        if (buf[i] == '\r') {
-                                buf[i] = 0;
-                                continue;
-                        }
-                        if (buf[i] == '\n') {
-                                buf[i] = 0;
-                                break;
-                        }
-                }
-                buf[size - 1] = 0;
-                if (strncmp(buf, "SSH-", 4) == 0)
-                        break;
-		if (j == 0)
-			strlcpy(firstline, buf, sizeof(firstline));
-        }
-	if (j >= MAXITER)
-		strlcpy(buf, firstline, sizeof(firstline));
-	else
-		atomicio(write, sock, SSHMAPVERSION, sizeof(SSHMAPVERSION));
-	close(sock);
-	return (0);
-}
-
-void
-waitforcommands(int readfd, int writefd)
-{
-	struct argument arg;
-	char buf[200];
-	char result[200];
-	struct sigaction sa;
-
-        /* block signals, get terminal modes and turn off echo */
-        memset(&sa, 0, sizeof(sa));
-        sa.sa_handler = sigalrm_handler;
-        (void) sigaction(SIGALRM, &sa, NULL);
-
-	while(atomicio(read, readfd, &arg, sizeof(arg))) {
-		/* exit command */
-		if (arg.a_ipv4.s_addr == 0)
-			return;
-
-		scanhost(&arg, result, sizeof(result));
-		ipv4toa(buf, sizeof(buf), &arg.a_ipv4);
-		strlcat(buf, " ", sizeof(buf));
-		strlcat(buf, result, sizeof(buf));
-
-		atomicio(write, writefd, buf, strlen(buf) + 1);
-	}
-}
-
 void
 printres(struct argument *exp, char *result)
 {
@@ -337,8 +217,9 @@ printres(struct argument *exp, char *result)
 }
 
 void
-scanips(struct argument *args, u_int32_t entries)
+scanips(struct argument *args, u_int32_t nargs)
 {
+	u_int32_t entries = 0;
 	fd_set *readset, *writeset;
 	int maxfd = 0, fdsetsz, counts = 0;
 	struct timeval tv, ptv;
@@ -360,7 +241,12 @@ scanips(struct argument *args, u_int32_t entries)
 	TAILQ_INIT(&readyqueue);
 	synlist_init();
 	timerclear(&ptv);
-	while (entries || counts || !synlist_empty()) {
+	while (TAILQ_FIRST(&genqueue) || entries ||
+	       counts || !synlist_empty()) {
+		/* Create new entries, if we need them */
+		if (!entries)
+			entries = populate(args, nargs);
+
 		memset(readset, 0, fdsetsz);
 		memset(writeset, 0, fdsetsz);
 		
@@ -598,162 +484,409 @@ ipv4toa(char *buf, size_t size, void *addr)
 	return 0;
 }
 
+void
+generate_free(struct generate *gen)
+{
+	struct address_node *node;
+
+	/* Remove generator and attached addr nodes */
+	for (node = TAILQ_FIRST(&gen->gen_anqueue);
+	     node;
+	     node = TAILQ_FIRST(&gen->gen_anqueue)) {
+		TAILQ_REMOVE(&gen->gen_anqueue,
+			     node, an_next);
+		xfree(node);
+	}
+
+	TAILQ_REMOVE(&genqueue, gen, gen_next);
+	xfree(gen);
+}
+
 /*
  * Given an IP prefix and mask create all addresses contained
  * exlcuding any addresses specified in the exlcude queues.
  */
 
 int
-populate(struct argument **args, u_int32_t *entries, sa_family_t type,
-	 void *address, int bits, int dorandom)
+populate(struct argument *args, int count)
 {
-	struct in_addr ipv4addr, ipv4next, ipv4end;
-	struct in_addr ipv4mask;
-	struct argument *work;
-	u_int32_t n, oldn, i;
+	struct generate *gen;
+	struct in_addr addr;
+	u_int32_t i = 0;
 
-	switch (type) {
-	case AF_INET:
-		ipv4addr = *(struct in_addr *)address;
-		ipv4mask.s_addr = 0xFFFFFFFF << (32 - bits);
-		ipv4addr.s_addr = ntohl(ipv4addr.s_addr);
-		ipv4addr.s_addr &= ipv4mask.s_addr;
-		ipv4end.s_addr = ipv4addr.s_addr | (~ipv4mask.s_addr);
-		break;
-		/* IPv6 should be here */
-	default:
-		fprintf(stderr, "Unknown address family: %d\n", type);
-		return -1;
-	}
+	while (TAILQ_FIRST(&genqueue) && count) {
+		gen = TAILQ_FIRST(&genqueue);
+		
+		/* Initalize generator */
+		if (!gen->gen_current) {
+			if (gen->gen_flags & FLAGS_USERANDOM)
+				rndsboxinit(gen->gen_seed);
 
-	if (!dorandom) {
-		n = 0;
-		for (ipv4next = ipv4addr;
-		     ipv4next.s_addr <= ipv4end.s_addr &&
-			     !(ipv4next.s_addr == INADDR_ANY && n > 0);
-		     ipv4next.s_addr++, n++) {
-			ipv4next = exclude(ipv4next, &excludequeue);
-			if (ipv4next.s_addr > ipv4end.s_addr)
-				break;
+			gen->gen_current = gen->gen_start;
 		}
-		if (n < (1 << (32 - bits)))
-			fprintf (stderr, "Excluding %d addresses\n",
-				 (1 << (32 - bits)) - n);
 
-		n += *entries;
-	} else {
-		struct in_addr tmp;
-		u_int32_t numbers;
-
-		rndsboxinit();
-		n = 0;
-		numbers = 1 << (32 - bits);
-		for (i = 0; (n < dorandom) && (i < numbers); i++) {
-			ipv4next = rndgetaddr(&ipv4addr, bits, i);
-			tmp = exclude(ipv4next, &rndexclqueue);
-			if (ipv4next.s_addr != tmp.s_addr)
-				continue;
-			if (rndexclude) {
-				tmp = exclude(ipv4next, &excludequeue);
-				if (ipv4next.s_addr != tmp.s_addr)
-					continue;
+		while (count) {
+			if (next_address(gen, &addr) == -1) {
+				generate_free(gen);
+				break;
 			}
 
-			n++;
-		}
-		if (n != dorandom)
-			fprintf(stderr, "Excluding %d addresses\n",
-				dorandom - n);
-		if (i > n)
-			fprintf(stderr, "Skipped %d addresses\n", i - n);
-		n += *entries;
-	}
-	
-	if (*args != NULL)
-		work = realloc(*args, n * sizeof(struct argument));
-	else
-		work = malloc(n * sizeof(struct argument));
-	if (work == NULL) {
-		warn("{Re,m}alloc failed: %d", n * sizeof(struct argument));
-		return -1;
-	}
-	
-	oldn = *entries;
-	memset(work + oldn, 0, (n - oldn) * sizeof(struct argument));
+			args[i].a_type = AF_INET;
+			args[i].a_ipv4.s_addr = htonl(addr.s_addr);
 
-	if (!dorandom) {
-		i = 0;
-		for (; ipv4addr.s_addr <= ipv4end.s_addr &&
-			     ipv4addr.s_addr != INADDR_ANY;
-		     ipv4addr.s_addr++, i++) {
-			ipv4addr = exclude(ipv4addr, &excludequeue); 
-			if (ipv4addr.s_addr > ipv4end.s_addr)
-				break;
-			work[i + oldn].a_type = type;
-			work[i + oldn].a_ipv4.s_addr = htonl(ipv4addr.s_addr);
-		}
-	} else {
-		int count = 0;
-		struct in_addr tmp;
-		u_int32_t numbers;
-
-		numbers = 1 << (32 - bits);
-		for (i = 0; (count < dorandom) && (i < numbers); i++) {
-			ipv4next = rndgetaddr(&ipv4addr, bits, i);
-			tmp = exclude(ipv4next, &rndexclqueue);
-			if (ipv4next.s_addr != tmp.s_addr)
-				continue;
-			if (rndexclude) {
-				tmp = exclude(ipv4next, &excludequeue);
-				if (ipv4next.s_addr != tmp.s_addr)
-					continue;
-			}
-			
-			work[count + oldn].a_type = type;
-			work[count + oldn].a_ipv4.s_addr = htonl(ipv4next.s_addr);
-			count++;
+			count--;
+			i++;
 		}
 	}
 
-	*args = work;
-	*entries = n;
-
-	return 0;
+	return (i);
 }
 
 int
-mix(struct argument **args, u_int32_t n)
+address_from_offset(struct address_node *an, u_int32_t offset,
+		    struct in_addr *addr)
 {
-	struct argument *work, *oldargs;
-	u_int32_t i, tmp;
+	for (; an; an = TAILQ_NEXT(an, an_next)) {
+		if (an->an_ipv4start.s_addr + offset <= an->an_ipv4end.s_addr)
+			break;
+		offset -= an->an_ipv4end.s_addr - an->an_ipv4start.s_addr + 1;
+	}
 
-	if ((work = malloc(n * sizeof(struct argument))) == NULL) {
-		warn("malloc failed");
+	if (an == NULL)
+		return (-1);
+
+	addr->s_addr = an->an_ipv4start.s_addr + offset;
+	
+	return (0);
+}
+
+/*
+ * get the next address, keep state.
+ */
+
+int
+next_address(struct generate *gen, struct in_addr *addr)
+{
+	struct in_addr ipv4addr, tmp;
+	u_int32_t offset;
+	int done = 0, random;
+
+	/* Check if generator has been exhausted */
+	if (gen->gen_n >= gen->gen_max)
+		return (-1);
+
+	random = gen->gen_flags & FLAGS_USERANDOM;
+
+	do {
+		/* Get offset into address range */
+		if (random)
+			offset = rndgetaddr(gen->gen_bits,
+					    gen->gen_current);
+		else
+			offset = gen->gen_current;
+		
+		gen->gen_current += gen->gen_iterate;
+		
+		if (address_from_offset(TAILQ_FIRST(&gen->gen_anqueue),
+					offset, &ipv4addr) == -1)
+			continue;
+		
+		if (!random || rndexclude) {
+			tmp = exclude(ipv4addr, &excludequeue);
+			if (ipv4addr.s_addr != tmp.s_addr) {
+				if (random)
+					continue;
+
+				/* In linear mode, we can skip these */
+				offset = gen->gen_current;
+				offset += tmp.s_addr - ipv4addr.s_addr;
+				if (offset < gen->gen_current) {
+					gen->gen_current = gen->gen_end;
+					break;
+				}
+				gen->gen_current = offset;
+				
+				if (gen->gen_iterate == 1)
+					continue;
+
+				/* Adjust for splits */
+				offset /= gen->gen_iterate;
+				offset *= gen->gen_iterate;
+
+				offset += gen->gen_start;
+
+				if (offset < gen->gen_current)
+					offset += gen->gen_iterate;
+				if (offset < gen->gen_current) {
+					gen->gen_current = gen->gen_end;
+					break;
+				}
+
+				gen->gen_current = offset;
+				continue;
+			}
+		}
+		
+		if (gen->gen_flags & FLAGS_USERANDOM) {
+			tmp = exclude(ipv4addr, &rndexclqueue);
+			if (ipv4addr.s_addr != tmp.s_addr) {
+				continue;
+			}
+		}
+		
+		/* We have an address */
+		done = 1;
+	} while ((gen->gen_current < gen->gen_end) && !done);
+
+	if (!done)
+		return (-1);
+
+	gen->gen_n += gen->gen_iterate;
+
+	*addr = ipv4addr;
+
+	return (0);
+}
+
+struct address_node *
+address_node_get(char *line)
+{
+	struct address_node *an;
+	struct in_addr ipv4mask;
+	int bits;
+
+	/* Allocate an address node */
+	an = xmalloc(sizeof(struct address_node));
+	memset(an, 0, sizeof(struct address_node));
+	if (parseaddress(line, &an->an_ipv4start, &bits) == -1) {
+		fprintf(stderr, "Can not parse %s\n", line);
+		return (NULL);
+	}
+
+	an->an_bits = bits;
+	an->an_type = AF_INET;
+
+	/* Calculate start and end address for this node */
+	if (bits > 0)
+		ipv4mask.s_addr = 0xFFFFFFFF << (32 - bits);
+	else
+		ipv4mask.s_addr = 0;
+
+	an->an_ipv4start.s_addr = ntohl(an->an_ipv4start.s_addr);
+	an->an_ipv4start.s_addr &= ipv4mask.s_addr;
+	an->an_ipv4end.s_addr = an->an_ipv4start.s_addr | (~ipv4mask.s_addr);
+
+	return (an);
+}
+
+/*
+ * Creates a generator from a command line
+ * [split(x/n)/][random(x,s)/][(]<address/mask> .... [)]
+ */
+
+int generate_split(struct generate *gen, char **pline)
+{
+	char *line, *end;
+
+	line = *pline;
+
+	if ((end = strstr(line, ")/")) == NULL ||
+	    strchr(line, '/') < end) {
+		fprintf(stderr, "Split not terminated correctly: %s\n", line);
 		return (-1);
 	}
 
-	oldargs = *args;
-	for (i = 0; i < n; i++) {
-		tmp = arc4random() % (n - i);
-		work[i] = oldargs[tmp];
-		oldargs[tmp] = oldargs[n - i - 1];
+	line = 	strsep(pline, "/");
+
+	/* Generate a random scan entry */
+	if (sscanf(line, "split(%d,%d)/",
+		   &gen->gen_start, &gen->gen_iterate) != 2)
+		return (-1);
+		
+	if (!gen->gen_start || gen->gen_start > gen->gen_iterate) {
+		fprintf(stderr, "Invalid start/iterate pair: %d/%d\n",
+			gen->gen_start, gen->gen_iterate);
+		return (-1);
 	}
 
-	*args = work;
-	free (oldargs);
+	/* Internally, we start counting at 0 */
+	gen->gen_start--;
 
 	return (0);
 }
 
+/*
+ * Creates a generator from a command line
+ * [split(x/n)/][random(x,s)/][(]<address/mask> .... [)]
+ */
+
+int generate_random(struct generate *gen, char **pline)
+{
+	int i;
+	char seed[31], *line, *end;
+
+	line = *pline;
+
+	if ((end = strstr(line, ")/")) == NULL ||
+	    strchr(line, '/') < end) {
+		fprintf(stderr, "Random not terminated correctly: %s\n", line);
+		return (-1);
+	}
+
+	line = strsep(pline, "/");
+
+	/* Generate a random scan entry */
+	seed[0] = '\0';
+	if (sscanf(line, "random(%d,%30s)/", &gen->gen_max, seed) < 1)
+		return (-1);
+		
+	/* Generate seed from string */
+	if (strlen(seed)) {
+		MD5_CTX ctx;
+		u_char digest[16];
+		u_int32_t *tmp = (u_int32_t *)digest;
+
+		MD5Init(&ctx);
+		MD5Update(&ctx, seed, strlen(seed));
+		MD5Final(digest, &ctx);
+
+		gen->gen_seed = 0;
+		for (i = 0; i < 4; i ++)
+			gen->gen_seed ^= *tmp++;
+				
+	} else
+		gen->gen_seed = arc4random();
+
+	gen->gen_flags |= FLAGS_USERANDOM;
+
+	return (0);
+}
+
+int
+generate(char *line)
+{
+	struct generate *gen;
+	struct address_node *an;
+	u_int32_t count, tmp;
+	char *p;
+	int bits, i, done;
+
+	gen = xmalloc(sizeof(struct generate));
+	memset(gen, 0, sizeof(struct generate));
+	TAILQ_INIT(&gen->gen_anqueue);
+
+	/* Insert in generator queue, on failure generate_free removes it */
+	TAILQ_INSERT_TAIL(&genqueue, gen, gen_next);
+
+	done = 0;
+	while (!done) {
+		done = 1;
+		if (strncmp(line, "random(", 7) == 0) {
+			if (gen->gen_flags & FLAGS_USERANDOM) {
+				fprintf(stderr,
+					"Random already specified: %s\n",
+					line);
+				goto fail;
+			}
+			if (generate_random(gen, &line) == -1)
+				goto fail;
+
+			done = 0;
+		} else if (strncmp(line, "split(", 6) == 0) {
+			if (gen->gen_iterate) {
+				fprintf(stderr,
+					"Split already specified: %s\n",
+					line);
+				goto fail;
+			}
+			if (generate_split(gen, &line) == -1)
+				goto fail;
+
+			done = 0;
+		}
+	}
+
+	/* If no special split is specified, always iterated by 1 */
+	if (!gen->gen_iterate)
+		gen->gen_iterate = 1;
+
+	if (line[0] == '(') {
+		char *end;
+		
+		line++;
+		if ((end = strchr(line, ')')) == NULL) {
+			fprintf(stderr, "Missing ')' in line: %s\n", line);
+			goto fail;
+		}
+		*end = '\0';
+		
+	}
+
+	while (line && (p = strsep(&line, " "))) {
+		if ((an = address_node_get(p)) == NULL)
+			goto fail;
+
+		TAILQ_INSERT_TAIL(&gen->gen_anqueue, an, an_next);
+	}
+
+	/* Try to find out the effective bit range */
+	count = 0;
+	for (an = TAILQ_FIRST(&gen->gen_anqueue); an;
+	     an = TAILQ_NEXT(an, an_next)) {
+		bits = an->an_bits;
+		if (bits == 0) {
+			count = -1;
+			break;
+		}
+
+		if (count + (1 << (32 - bits)) < count) {
+			count = -1;
+			break;
+		}
+
+		count += 1 << (32 - bits);
+	}
+
+	/* Try to convert count into a network mask */
+	bits = 0;
+	tmp = count;
+	for (i = -1; tmp; tmp >>= 1, i++) {
+		if (tmp & 1)
+			bits++;
+	}
+
+	/* a count of 01100, results in bits = 29, but it should be 28 */
+	gen->gen_bits = 32 - i;
+	if (bits > 1)
+		gen->gen_bits--;
+	bits = gen->gen_bits;
+
+	if (gen->gen_flags & FLAGS_USERANDOM) {
+		if (bits == 0)
+			gen->gen_end = -1;
+		else 
+			gen->gen_end = 1 << (32 - bits);
+	} else
+		gen->gen_end = count;
+
+	if (gen->gen_max == 0)
+		gen->gen_max = count;
+
+	return (0);
+ fail:
+	if (gen)
+		generate_free(gen);
+
+	return (-1);
+}
+
 int
 main(int argc, char **argv)
 {
 	char *name, *src = NULL;
 	int ch;
-	u_int32_t entries = 0;
 	struct argument *args = NULL;
-	struct in_addr address;
-	int bits, dorandom;
+	u_int32_t nargs;
 	int failonexclude = 1;
 
 	name = argv[0];
@@ -791,7 +924,9 @@ main(int argc, char **argv)
 
 
 	/* revoke privs */
+#ifdef HAVE_SETEUID
         seteuid(getuid());
+#endif /* HAVE_SETEUID */
         setuid(getuid());
 
 	setupchildren();
@@ -799,30 +934,36 @@ main(int argc, char **argv)
 	if (setupexcludes() == -1 && failonexclude)
 		exit(1);
 
+	nargs = EXPANDEDARGS;
+	args = xmalloc(nargs * sizeof(struct argument));
+	memset(args, 0, nargs * sizeof(struct argument)); 
+
+	TAILQ_INIT(&genqueue);
+
 	while (argc) {
-		if (sscanf(argv[0], "random(%d)/", &dorandom) == 1) {
-			strsep(&argv[0], "/");
-		} else
-			dorandom = 0;
-		if (parseaddress(argv[0], &address, &bits) == -1) {
-			fprintf(stderr, "Can not parse %s\n", argv[0]);
-			exit (1);
-		}
-		if (populate(&args, &entries, AF_INET, &address, bits,
-			     dorandom) == -1)
-			errx(1, "populate failed");
+		if (generate(argv[0]) == -1)
+			warnx("generate failed on %s", argv[0]);
 
 		argv++;
 		argc--;
 	}
+	/*
+	{
+		int i, n;
+		struct argument args[100];
 
-	if (entries == 0)
+		while ((n = populate(args, 100)) != 0) {
+			for (i = 0; i < n; i++)
+				printres(&args[i], "good");
+		}
+
+		exit(1);
+	}
+	*/
+	if (!TAILQ_FIRST(&genqueue))
 		errx(1, "nothing to scan");
 
-	if (mix(&args, entries) == -1)
-		exit (1);
-
-	scanips(args, entries);
+	scanips(args, nargs);
 
 	return (1);
 }
